/**
 *    Copyright (C) 2021 Graham Leggett <minfrin@sharp.fm>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*
 * redwax_openssl - OpenSSL routines for munching certificates
 *
 */

#include <apr_crypto.h>
#include <apr_lib.h>
#include <apr_strings.h>

#include "config.h"
#include "redwax-tool.h"

#if HAVE_OPENSSL_PEM_H

#include <openssl/asn1.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/pkcs12.h>
#include <openssl/rsa.h>
#include <openssl/ui.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#if HAVE_OPENSSL_CORE_NAMES_H
#include <openssl/core_names.h>
#endif

#define REDWAX_OPENSSL_SEARCH "search"
#define REDWAX_OPENSSL_VERIFY "verify"

#define REDWAX_PKCS12_MIN 8
#define REDWAX_PKCS12_MAX HUGE_STRING_LEN

static STACK_OF(X509) *cert_index;
static STACK_OF(X509) *chain_index;
static STACK_OF(X509) *trusted_index;
static STACK_OF(X509_CRL) *crl_index;

static int redwax_x509_idx = -1;
static int redwax_x509_store_ctx_idx = -1;

static int redwax_get_x509_index()
{
    if (redwax_x509_idx == -1) {
        redwax_x509_idx =
                    X509_get_ex_new_index(0,
                            "Application Data for X509",
                            NULL, NULL, NULL);
    }
    return redwax_x509_idx;
}

static int redwax_get_x509_store_ctx_index()
{
    if (redwax_x509_store_ctx_idx == -1) {
        redwax_x509_store_ctx_idx =
                    X509_STORE_CTX_get_ex_new_index(0,
                            "Application Data for X509_STORE_CTX",
                            NULL, NULL, NULL);
    }
    return redwax_x509_store_ctx_idx;
}

#if !HAVE_X509_STORE_CTX_GET_NUM_UNTRUSTED
int X509_STORE_CTX_get_num_untrusted(X509_STORE_CTX *ctx)
{
    return ctx->last_untrusted;
}
#endif

#if !HAVE_X509_STORE_GET0_PARAM
X509_VERIFY_PARAM *X509_STORE_get0_param(X509_STORE *ctx)
{
    return ctx->param;
}
#endif

#if !HAVE_X509_STORE_CTX_SET0_TRUSTED_STACK
#define X509_STORE_CTX_set0_trusted_stack X509_STORE_CTX_trusted_stack
#endif

#if !HAVE_X509_GET0_NOTBEFORE
#define X509_get0_notBefore X509_get_notBefore
#endif

#if !HAVE_X509_GET0_NOTAFTER
#define X509_get0_notAfter X509_get_notAfter
#endif

#if !HAVE_X509_GET_EXTENSION_FLAGS
uint32_t X509_get_extension_flags(X509 *x)
{
    /* Call for side-effect of computing hash and caching extensions */
    X509_check_purpose(x, -1, 0);
    return x->ex_flags;
}
#endif

#if !HAVE_EVP_PKEY_GET_BN_PARAM

#if !HAVE_RSA_GET0_N
const BIGNUM *RSA_get0_n(const RSA *r)
{
    return r->n;
}
#endif

#if !HAVE_RSA_GET0_E
const BIGNUM *RSA_get0_e(const RSA *r)
{
    return r->e;
}
#endif

#if !HAVE_RSA_GET0_D
const BIGNUM *RSA_get0_d(const RSA *r)
{
    return r->d;
}
#endif

#if !HAVE_RSA_GET0_P
const BIGNUM *RSA_get0_p(const RSA *r)
{
    return r->p;
}
#endif

#if !HAVE_RSA_GET0_Q
const BIGNUM *RSA_get0_q(const RSA *r)
{
    return r->q;
}
#endif

#if !HAVE_RSA_GET0_DMP1
const BIGNUM *RSA_get0_dmp1(const RSA *r)
{
    return r->dmp1;
}
#endif

#if !HAVE_RSA_GET0_DMQ1
const BIGNUM *RSA_get0_dmq1(const RSA *r)
{
    return r->dmq1;
}
#endif

#if !HAVE_RSA_GET0_IQMP
const BIGNUM *RSA_get0_iqmp(const RSA *r)
{
    return r->iqmp;
}
#endif

#endif

static void redwax_openssl_print_errors(redwax_tool_t *r)
{
    if (!r->quiet && !r->complete) {
        ERR_print_errors_fp(stderr);
    }
    else {
        ERR_clear_error();
    }
}

static const char *redwax_openssl_name(apr_pool_t *p, X509_NAME *name)
{
    BIO *bio;
    char *buf = NULL;
    int len = 0;

    if ((bio = BIO_new(BIO_s_mem())) == NULL) {
        return NULL;
    }

    if (name != NULL) {
        X509_NAME_print_ex(bio,
                        name,
                        0, XN_FLAG_RFC2253);
        len = BIO_get_mem_data(bio, &buf);
    }

    buf = apr_pstrndup(p, buf, len);

    BIO_free(bio);

    return buf;
}

static apr_status_t cleanup_alloc(void *dummy)
{
    if (dummy) {
        OPENSSL_free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_bio(void *dummy)
{
    if (dummy) {
        free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_x509(void *dummy)
{
    if (dummy) {
        X509_free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_pkcs12(void *dummy)
{
    if (dummy) {
        PKCS12_free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_evp_pkey(void *dummy)
{
    if (dummy) {
        EVP_PKEY_free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_p8inf(void *dummy)
{
    if (dummy) {
        PKCS8_PRIV_KEY_INFO_free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_ui(void *dummy)
{
    if (dummy) {
        UI_free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_sk_x509(void *dummy)
{
    if (dummy) {
        sk_X509_pop_free(dummy, X509_free);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_sk_x509_crl(void *dummy)
{
    if (dummy) {
        sk_X509_CRL_pop_free(dummy, X509_CRL_free);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_x509_store(void *dummy)
{
    if (dummy) {
        X509_STORE_free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t cleanup_x509_store_ctx(void *dummy)
{
    if (dummy) {
        X509_STORE_CTX_free(dummy);
    }

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_initialise(redwax_tool_t *r)
{
    cert_index = sk_X509_new_null();
    chain_index = sk_X509_new_null();
    trusted_index = sk_X509_new_null();
    crl_index = sk_X509_CRL_new_null();

    apr_pool_cleanup_register(r->pool, cert_index, cleanup_sk_x509,
            apr_pool_cleanup_null);
    apr_pool_cleanup_register(r->pool, chain_index, cleanup_sk_x509,
            apr_pool_cleanup_null);
    apr_pool_cleanup_register(r->pool, trusted_index, cleanup_sk_x509,
            apr_pool_cleanup_null);
    apr_pool_cleanup_register(r->pool, crl_index, cleanup_sk_x509_crl,
            apr_pool_cleanup_null);

    redwax_get_x509_index();
    redwax_get_x509_store_ctx_index();

    return OK;
}

static apr_status_t redwax_openssl_process_pem_in(redwax_tool_t *r,
        const char *file, const char *secret)
{

    BIO *bio;
    char *name = NULL, *header = NULL;
    unsigned char *data = NULL;

    long len, error = 0;

    if (!strcmp(file, "-")) {
        if (r->complete) {
            return APR_ENOENT;
        }
        if ((bio = BIO_new_fp(stdin, BIO_NOCLOSE)) == NULL) {
            redwax_openssl_print_errors(r);
            return APR_ENOMEM;
        }
    }
    else if ((bio = BIO_new(BIO_s_file())) == NULL) {
        redwax_openssl_print_errors(r);
        return APR_ENOMEM;
    }
    else if (BIO_read_filename(bio, file) <= 0) {
        redwax_openssl_print_errors(r);
        BIO_free(bio);
        return APR_ENOENT;
    }

    apr_pool_cleanup_register(r->pool, bio, cleanup_bio,
            apr_pool_cleanup_null);


    for (;;) {

        const unsigned char *der;

        PKCS8_PRIV_KEY_INFO *p8inf = NULL;
        EVP_PKEY *pkey = NULL;

        /* header is deprecated, for details see PEM_get_EVP_CIPHER_INFO
         * and PEM_do_header in OpenSSL. We do not (yet) support it here.
         */

        if (!PEM_read_bio(bio, &name, &header, &data, &len)) {
            error = ERR_GET_REASON(ERR_peek_last_error());
            if (error == PEM_R_NO_START_LINE) {
                ERR_clear_error();
            }
            break;
        }

        apr_pool_cleanup_register(r->pool, name, cleanup_alloc,
                apr_pool_cleanup_null);
        apr_pool_cleanup_register(r->pool, header, cleanup_alloc,
                apr_pool_cleanup_null);
        apr_pool_cleanup_register(r->pool, data, cleanup_alloc,
                apr_pool_cleanup_null);

        der = data;

        /* is this a duplicate? if so, skip */
        if (apr_hash_get(r->duplicates_index, der, len)) {
            continue;
        }
        else {
            apr_hash_set(r->duplicates_index, der, len, der);
        }

        if ((strcmp(name, PEM_STRING_X509) == 0) ||
            (strcmp(name, PEM_STRING_X509_OLD) == 0)) {

            redwax_certificate_t *cert;

            X509 *x = d2i_X509(NULL, &der, len);

            if (!x) {
                redwax_print_error(r, "Could not read certificate from '%s', skipping.\n",
                        file);
                redwax_openssl_print_errors(r);
                continue;
            }

            if (X509_check_ca(x)) {

                cert = apr_array_push(r->intermediates_in);

                apr_pool_create(&cert->pool, r->pool);

                cert->common.type = REDWAX_CERTIFICATE_X509;
                cert->common.category = REDWAX_CERTIFICATE_INTERMEDIATE;

                redwax_print_error(r, "pem-in: intermediate: %s\n",
                        redwax_openssl_name(cert->pool,
                                X509_get_subject_name(x)));
            }
            else {

                cert = apr_array_push(r->certs_in);

                apr_pool_create(&cert->pool, r->pool);

                cert->common.type = REDWAX_CERTIFICATE_X509;
                cert->common.category = REDWAX_CERTIFICATE_END_ENTITY;

                redwax_print_error(r, "pem-in: certificate: %s\n",
                        redwax_openssl_name(cert->pool,
                                X509_get_subject_name(x)));

            }

            cert->header = header;
            cert->der = data;
            cert->len = len;

            rt_run_normalise_certificate(r, cert, 1);

        }

        else if ((strcmp(name, PEM_STRING_X509_TRUSTED) == 0)) {

            redwax_certificate_t *cert;

            X509 *x = d2i_X509_AUX(NULL, &der, len);

            if (!x) {
                redwax_print_error(r, "Could not read certificate from '%s', skipping.\n",
                        file);
                redwax_openssl_print_errors(r);
                continue;
            }

            else {

                unsigned char *der = NULL;

                cert = apr_array_push(r->trusted_in);

                apr_pool_create(&cert->pool, r->pool);

                cert->common.type = REDWAX_CERTIFICATE_X509;
                cert->common.category = REDWAX_CERTIFICATE_TRUSTED;

                cert->header = header;
                cert->name = (const char *)X509_alias_get0(x, NULL);

                cert->len = i2d_X509(x, &der);
                cert->der = der;

                rt_run_normalise_certificate(r, cert, 1);

                redwax_print_error(r, "pem-in: trusted: %s\n",
                        redwax_openssl_name(cert->pool,
                                X509_get_subject_name(x)));

            }
        }

        else if (strcmp(name, PEM_STRING_X509_CRL) == 0) {

            redwax_crl_t *crl;

            X509_CRL *c = d2i_X509_CRL(NULL, &der, len);

            if (c) {

                redwax_print_error(r, "pem-in: crl: %s\n",
                        redwax_openssl_name(r->pool, X509_CRL_get_issuer(c)));

                sk_X509_CRL_push(crl_index, c);

                crl = apr_array_push(r->crls_in);

                apr_pool_create(&crl->pool, r->pool);

                crl->header = header;
                crl->der = data;
                crl->len = len;

//                 X509_CRL_free(c);
             }

        }

        else if (strcmp(name, PEM_STRING_RSA) == 0) {

            if (!(pkey = d2i_PrivateKey(EVP_PKEY_RSA, NULL, &der, len)) ||
                    !(p8inf = EVP_PKEY2PKCS8(pkey))) {

                redwax_print_error(r, "Could not read RSA private key from '%s', skipping.\n",
                        file);
                redwax_openssl_print_errors(r);
            }

        }

        else if (strcmp(name, PEM_STRING_DSA) == 0) {

            if (!(pkey = d2i_PrivateKey(EVP_PKEY_DSA, NULL, &der, len)) ||
                    !(p8inf = EVP_PKEY2PKCS8(pkey))) {

                redwax_print_error(r, "Could not read DSA private key from '%s', skipping.\n",
                        file);
                redwax_openssl_print_errors(r);
            }

        }

        else if (strcmp(name, PEM_STRING_ECPRIVATEKEY) == 0) {

            if (!(pkey = d2i_PrivateKey(EVP_PKEY_EC, NULL, &der, len)) ||
                    !(p8inf = EVP_PKEY2PKCS8(pkey))) {

                redwax_print_error(r, "Could not read EC private key from '%s', skipping.\n",
                        file);
                redwax_openssl_print_errors(r);
            }

        }
        else if (strcmp(name, PEM_STRING_PKCS8INF) == 0) {

            BIO *kbio;

            if ((kbio = BIO_new_mem_buf(der, len)) == NULL) {
                return APR_ENOMEM;
            }

            if (!(p8inf = d2i_PKCS8_PRIV_KEY_INFO_bio(kbio, NULL)) ||
                    !(pkey = EVP_PKCS82PKEY(p8inf))) {

                redwax_print_error(r, "Could not read private key from '%s', skipping.\n",
                        file);
                redwax_openssl_print_errors(r);
            }

            BIO_free(kbio);
        }

        if (p8inf) {
            apr_pool_cleanup_register(r->pool, p8inf, cleanup_p8inf,
                    apr_pool_cleanup_null);
        }

        if (pkey) {
            apr_pool_cleanup_register(r->pool, pkey, cleanup_evp_pkey,
                    apr_pool_cleanup_null);
        }

        /* handle keys */
        if (p8inf && pkey) {

            redwax_key_t *key;

            BIO *kbio;

#if HAVE_EVP_PKEY_GET0_DESCRIPTION
            redwax_print_error(r, "pem-in: private key: %s\n",
                    EVP_PKEY_get0_description(pkey));
#else
            redwax_print_error(r, "pem-in: private key\n");
#endif

            if ((kbio = BIO_new(BIO_s_mem())) == NULL) {
                return APR_ENOMEM;
            }

            apr_pool_cleanup_register(r->pool, kbio, cleanup_bio,
                    apr_pool_cleanup_null);

            i2d_PKCS8_PRIV_KEY_INFO_bio(kbio, p8inf);

            key = apr_array_push(r->keys_in);

            apr_pool_create(&key->pool, r->pool);

            switch(EVP_PKEY_base_id(pkey)) {
            case EVP_PKEY_RSA: {
                key->common.type = REDWAX_KEY_RSA;
                break;
            }
            }

            key->header = header;
            key->len = BIO_get_mem_data(kbio, &key->der);

            rt_run_normalise_key(r, key, 1);

        }

    }

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_set_verify_param(redwax_tool_t *r, const char *arg)
{
    const X509_VERIFY_PARAM *param;

    if (!(param = X509_VERIFY_PARAM_lookup(arg))) {
        redwax_print_error(r,
                "Verify parameter not found: %s\n", arg);
        return APR_ENOENT;
    }

    r->verify_param = arg;

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_complete_verify_param(redwax_tool_t *r,
        apr_hash_t *params)
{
    int i;
    int count = X509_VERIFY_PARAM_get_count();

    for (i = 0; i < count; i++) {
        const char *name =
                X509_VERIFY_PARAM_get0_name(X509_VERIFY_PARAM_get0(i));

        apr_hash_set(params, name, APR_HASH_KEY_STRING, name);
    }

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_complete_filter_search(redwax_tool_t *r,
        apr_hash_t *filters)
{
    apr_hash_set(filters, REDWAX_OPENSSL_SEARCH,
            strlen(REDWAX_OPENSSL_SEARCH), REDWAX_OPENSSL_SEARCH);

    return APR_SUCCESS;
}

static int search_cert_match(redwax_tool_t *r, const redwax_certificate_t *cert)
{
    const unsigned char *der = cert->der;

    X509 *x = d2i_X509(NULL, &der, cert->len);

    int match = 0;

    if (x) {

        apr_hash_index_t *hi;
        const void *key;
        apr_ssize_t klen;

        for (hi = apr_hash_first(r->pool, r->emails); hi; hi = apr_hash_next(hi)) {
            apr_hash_this(hi, &key, &klen, NULL);

            if (X509_check_email(x, key, klen, 0) == 1) {
                match = 1;
            }

        }

        for (hi = apr_hash_first(r->pool, r->hostnames); hi; hi = apr_hash_next(hi)) {
            apr_hash_this(hi, &key, &klen, NULL);

             if (X509_check_host(x, key, klen, 0, NULL) == 1) {
                match = 1;
            }

        }

        for (hi = apr_hash_first(r->pool, r->ips); hi; hi = apr_hash_next(hi)) {
            apr_hash_this(hi, &key, &klen, NULL);

             if (X509_check_ip_asc(x, key, 0) == 1) {
                match = 1;
            }

        }

        X509_free(x);
    }

    return match;
}

static apr_status_t redwax_openssl_process_filter_search(redwax_tool_t *r,
        const char *arg)
{
    int i;

    if (strcmp(arg, REDWAX_OPENSSL_SEARCH)) {
        return DECLINED;
    }

    for (i = 0; i < r->certs_in->nelts; i++)
    {
        const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->certs_in, i,
                const redwax_certificate_t);

        if (search_cert_match(r, cert)) {

            redwax_certificate_t *ncert;
            redwax_certificate_t tcert = { 0 };

            tcert.header = cert->header;
            tcert.der = cert->der;
            tcert.len = cert->len;

            if (r->top && r->certs_out->nelts) {

                const redwax_certificate_t *ocert =
                        &APR_ARRAY_IDX(r->certs_out, 0,
                                const redwax_certificate_t);

                int diff = rt_run_compare_certificate(r, ocert, &tcert);

                /* no compare module, give up */
                if (diff == DECLINED)  {
                    redwax_print_error(r,
                            "When searching, there was no implementation "
                            "to compare certificates. Giving up.\n");
                    return APR_ENOENT;
                }

                /* new cert is better, blow away previous results */
                else if (diff == RIGHT) {
                    apr_array_clear(r->certs_out);
                    apr_array_clear(r->intermediates_out);
                    apr_array_clear(r->trusted_out);
                    apr_array_clear(r->keys_out);
                }

                /* original cert is better, ignore the new cert */
                else {
                    continue;
                }

            }

            ncert = apr_array_push(r->certs_out);

            memcpy(ncert, &tcert, sizeof(redwax_certificate_t));

            rt_run_search_chain(r, cert);
            rt_run_search_key(r, cert);
        }

    }

    for (i = 0; i < r->intermediates_in->nelts; i++)
    {
        const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->intermediates_in, i,
                const redwax_certificate_t);

        if (search_cert_match(r, cert)) {

            redwax_certificate_t *ncert = apr_array_push(r->intermediates_out);

            ncert->header = cert->header;
            ncert->der = cert->der;
            ncert->len = cert->len;

            rt_run_search_chain(r, cert);
            rt_run_search_key(r, cert);
         }

    }

    for (i = 0; i < r->trusted_in->nelts; i++)
    {
        const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->trusted_in, i,
                const redwax_certificate_t);

        if (search_cert_match(r, cert)) {

            redwax_certificate_t *ncert = apr_array_push(r->trusted_out);

            ncert->header = cert->header;
            ncert->der = cert->der;
            ncert->len = cert->len;

            rt_run_search_chain(r, cert);
            rt_run_search_key(r, cert);
         }

    }

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_complete_filter_verify(redwax_tool_t *r,
        apr_hash_t *filters)
{
    apr_hash_set(filters, REDWAX_OPENSSL_VERIFY,
            strlen(REDWAX_OPENSSL_VERIFY), REDWAX_OPENSSL_VERIFY);

    return APR_SUCCESS;
}

static int verify_cb(int ok, X509_STORE_CTX *ctx)
{
    BIO *bio;
    X509 *current_cert = X509_STORE_CTX_get_current_cert(ctx);
    char *buf = NULL;
    int len = 0;

    redwax_tool_t *r =
             X509_STORE_CTX_get_ex_data(ctx,
                     redwax_get_x509_store_ctx_index());

    if ((bio = BIO_new(BIO_s_mem())) == NULL) {
        return 0;
    }

    if (current_cert != NULL) {
        X509_NAME_print_ex(bio,
                        X509_get_subject_name(current_cert),
                        0, XN_FLAG_RFC2253);
        len = BIO_get_mem_data(bio, &buf);
    }

    if (!ok) {
        redwax_print_error(r,
                "verify-filter: %d: %.*s: verify failed: %s\n",
                X509_STORE_CTX_get_error_depth(ctx), len, buf,
                X509_verify_cert_error_string(
                        X509_STORE_CTX_get_error(ctx)));
    }
    else {
        redwax_print_error(r,
                "verify-filter: %d: %.*s: verify ok\n",
                X509_STORE_CTX_get_error_depth(ctx), len, buf);
    }

    BIO_free(bio);

    return ok;
}

static apr_status_t redwax_openssl_process_filter_verify(redwax_tool_t *r,
        const char *arg)
{
    X509_STORE_CTX *ctx;
    X509_STORE *store;
    apr_hash_index_t *hi;
    const void *key;
    apr_ssize_t klen;
    int i, j;

    if (strcmp(arg, REDWAX_OPENSSL_VERIFY)) {
        return DECLINED;
    }

    store = X509_STORE_new();

    apr_pool_cleanup_register(r->pool, store, cleanup_x509_store,
            apr_pool_cleanup_null);

    ctx = X509_STORE_CTX_new();

    apr_pool_cleanup_register(r->pool, ctx, cleanup_x509_store_ctx,
            apr_pool_cleanup_null);

    if (r->verify_param) {
        const X509_VERIFY_PARAM *param;

        if (!(param = X509_VERIFY_PARAM_lookup(r->verify_param))) {
            redwax_print_error(r,
                    "Verify parameter not found: %s", r->verify_param);
            return APR_ENOENT;
        }

        X509_VERIFY_PARAM_inherit(X509_STORE_get0_param(store), param);
    }

    if (apr_hash_count(r->emails) <= 1) {
        for (hi = apr_hash_first(r->pool, r->emails); hi; hi = apr_hash_next(hi)) {
            apr_hash_this(hi, &key, &klen, NULL);

            X509_VERIFY_PARAM_set1_email(X509_STORE_get0_param(store),
                    (const char *)key, klen);
        }
    }
    else {
        redwax_print_error(r,
                "When verifying, email address can only be specified once\n");
        return APR_ENOENT;
    }

    for (hi = apr_hash_first(r->pool, r->hostnames); hi; hi = apr_hash_next(hi)) {
        apr_hash_this(hi, &key, &klen, NULL);

        X509_VERIFY_PARAM_add1_host(X509_STORE_get0_param(store),
                (const char *)key, klen);
    }

    if (apr_hash_count(r->ips) <= 1) {
        for (hi = apr_hash_first(r->pool, r->ips); hi; hi = apr_hash_next(hi)) {
            apr_hash_this(hi, &key, &klen, NULL);

            X509_VERIFY_PARAM_set1_ip_asc(X509_STORE_get0_param(store),
                    (const char *)key);
        }
    }
    else {
        redwax_print_error(r,
                "When verifying, ip address can only be specified once\n");
        return APR_ENOENT;
    }


    for (i = 0; i < r->certs_in->nelts; i++)
    {
        const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->certs_in, i,
                const redwax_certificate_t);

        const unsigned char *der = cert->der;

         X509 *x = d2i_X509(NULL, &der, cert->len);
         if (x) {

             if (!X509_STORE_CTX_init(ctx, store, x, chain_index)) {
                 return APR_ENOMEM;
             }

             X509_STORE_CTX_set_ex_data(ctx,
                     redwax_get_x509_store_ctx_index(), (void *)r);

             X509_STORE_CTX_set0_trusted_stack(ctx, trusted_index);
             X509_STORE_CTX_set0_crls(ctx, crl_index);

             X509_STORE_CTX_set_verify_cb(ctx, &verify_cb);

             if (X509_verify_cert(ctx) > 0 &&
                     X509_STORE_CTX_get_error(ctx) == X509_V_OK) {

                 redwax_certificate_t *ncert;
                 redwax_certificate_t tcert = { 0 };

                 tcert.header = cert->header;
                 tcert.der = cert->der;
                 tcert.len = cert->len;

                 if (r->top && r->certs_out->nelts) {

                     const redwax_certificate_t *ocert =
                             &APR_ARRAY_IDX(r->certs_out, 0,
                                     const redwax_certificate_t);

                     int diff = rt_run_compare_certificate(r, ocert, &tcert);

                     /* no compare module, give up */
                     if (diff == DECLINED)  {
                         redwax_print_error(r,
                                 "When verifying, there was no implementation "
                                 "to compare certificates. Giving up.\n");
                         return APR_ENOENT;
                     }

                     /* new cert is better, blow away previous results */
                     else if (diff == RIGHT) {
                         apr_array_clear(r->certs_out);
                         apr_array_clear(r->intermediates_out);
                         apr_array_clear(r->trusted_out);
                         apr_array_clear(r->keys_out);
                     }

                     /* original cert is better, ignore the new cert */
                     else {
                         continue;
                     }

                 }

                 ncert = apr_array_push(r->certs_out);

                 memcpy(ncert, &tcert, sizeof(redwax_certificate_t));

                 rt_run_search_key(r, cert);

                 STACK_OF(X509) *xs = X509_STORE_CTX_get1_chain(ctx);
                 int num_untrusted = X509_STORE_CTX_get_num_untrusted(ctx);

                 for (j = 0; j < sk_X509_num(xs); j++) {
                     X509 *x = sk_X509_value(xs, j);

                     redwax_certificate_t *chain =
                             X509_get_ex_data(x, redwax_get_x509_index());

                     if (!chain) {

                         /* leaf cert, we're already handled above */

                     }
                     else if (j < num_untrusted) {

                         ncert = apr_array_push(r->intermediates_out);

                         ncert->header = chain->header;
                         ncert->der = chain->der;
                         ncert->len = chain->len;

                         rt_run_search_key(r, chain);
                     }
                     else {

                         ncert = apr_array_push(r->trusted_out);

                         ncert->header = chain->header;
                         ncert->der = chain->der;
                         ncert->len = chain->len;

                         rt_run_search_key(r, chain);
                     }

                 }
                 sk_X509_pop_free(xs, X509_free);

             }
             else {

                 /* verification failed */

             }

             X509_STORE_CTX_cleanup(ctx);
             X509_free(x);
         }

    }

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_process_pem_out(redwax_tool_t *r,
        const char *file, const char *secret)
{

    BIO *bio;
    int i;

    if (!strcmp(file, "-")) {
        if ((bio = BIO_new_fp(stdout, BIO_NOCLOSE)) == NULL) {
            redwax_openssl_print_errors(r);
            return APR_ENOMEM;
        }
    }
    else if ((bio = BIO_new(BIO_s_file())) == NULL) {
        redwax_openssl_print_errors(r);
        return APR_ENOMEM;
    }
    else if (BIO_write_filename(bio, (char *)file) <= 0) {
        redwax_openssl_print_errors(r);
        BIO_free(bio);
        return APR_ENOENT;
    }

    apr_pool_cleanup_register(r->pool, bio, cleanup_bio,
            apr_pool_cleanup_null);

    if (r->cert_out) {
        for (i = 0; i < r->certs_out->nelts; i++)
        {
            const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->certs_out, i, const redwax_certificate_t);

            const unsigned char *der = cert->der;

             X509 *x = d2i_X509(NULL, &der, cert->len);

             if (x) {

                 redwax_print_error(r, "pem-out: certificate: %s\n",
                         redwax_openssl_name(r->pool, X509_get_subject_name(x)));

                 if ((r->text && !X509_print_ex(bio, x, 0, 0)) ||
                         !PEM_write_bio_X509(bio, x)) {
                     redwax_openssl_print_errors(r);
                     X509_free(x);
                     return APR_ENOENT;
                 }
                 X509_free(x);
             }

        }
    }

    if (r->chain_out) {
        for (i = 0; i < r->intermediates_out->nelts; i++)
        {
            const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->intermediates_out, i, const redwax_certificate_t);

            const unsigned char *der = cert->der;

            X509 *x = d2i_X509(NULL, &der, cert->len);

            if (x) {

                redwax_print_error(r, "pem-out: intermediate: %s\n",
                        redwax_openssl_name(r->pool, X509_get_subject_name(x)));

                if ((r->text && !X509_print_ex(bio, x, 0, 0)) ||
                        !PEM_write_bio_X509(bio, x)) {
                    redwax_openssl_print_errors(r);
                    X509_free(x);
                    return APR_ENOENT;
                }
                X509_free(x);
            }

        }
    }

    if (r->root_out || r->trust_out) {
        for (i = 0; i < r->trusted_out->nelts; i++)
        {
            const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->trusted_out, i, const redwax_certificate_t);

            const unsigned char *der = cert->der;

            X509 *x = d2i_X509_AUX(NULL, &der, cert->len);

            if (x) {

                redwax_print_error(r, "pem-out: trusted: %s\n",
                        redwax_openssl_name(r->pool, X509_get_subject_name(x)));

                if (r->trust_out) {

                    if ((r->text && !X509_print_ex(bio, x, 0, 0)) ||
                            !PEM_write_bio_X509_AUX(bio, x)) {
                        redwax_openssl_print_errors(r);
                        X509_free(x);
                        return APR_ENOENT;
                    }

                }
                else {

                    if ((r->text && !X509_print_ex(bio, x, 0, 0)) ||
                            !PEM_write_bio_X509(bio, x)) {
                        redwax_openssl_print_errors(r);
                        X509_free(x);
                        return APR_ENOENT;
                    }

                }
                X509_free(x);
            }

        }
    }

    if (r->crl_out) {
        for (i = 0; i < r->crls_out->nelts; i++)
        {
            const redwax_crl_t *crl = &APR_ARRAY_IDX(r->crls_out, i, const redwax_crl_t);

            const unsigned char *der = crl->der;

            X509_CRL *c = d2i_X509_CRL(NULL, &der, crl->len);

            if (c) {

                redwax_print_error(r, "pem-out: crl: %s\n",
                        redwax_openssl_name(r->pool, X509_CRL_get_issuer(c)));

                if ((r->text && !X509_CRL_print(bio, c)) ||
                        !PEM_write_bio_X509_CRL(bio, c)) {
                    redwax_openssl_print_errors(r);
                    X509_CRL_free(c);
                    return APR_ENOENT;
                }
                   X509_CRL_free(c);
            }

        }
    }

    if (r->key_out) {
        for (i = 0; i < r->keys_out->nelts; i++)
        {
            const redwax_key_t *key = &APR_ARRAY_IDX(r->keys_out, i, const redwax_key_t);

            BIO *kbio;
            PKCS8_PRIV_KEY_INFO *p8inf;
            EVP_PKEY *pkey;

            if ((kbio = BIO_new_mem_buf(key->der, key->len)) == NULL) {
                return APR_ENOMEM;
            }

            apr_pool_cleanup_register(r->pool, kbio, cleanup_bio,
                    apr_pool_cleanup_null);

            if (!(p8inf = d2i_PKCS8_PRIV_KEY_INFO_bio(kbio, NULL)) ||
                    !(pkey = EVP_PKCS82PKEY(p8inf))) {

                redwax_openssl_print_errors(r);
                return APR_ENOENT;
            }

#if HAVE_EVP_PKEY_GET0_DESCRIPTION
            redwax_print_error(r, "pem-out: private key: %s\n",
                    EVP_PKEY_get0_description(pkey));
#else
            redwax_print_error(r, "pem-out: private key\n");
#endif

            if ((r->text && !EVP_PKEY_print_private(bio, pkey, 0, NULL)) ||
                    !PEM_write_bio_PKCS8_PRIV_KEY_INFO(bio, p8inf)) {

                redwax_openssl_print_errors(r);
                return APR_ENOENT;
            }

        }
    }

    return APR_SUCCESS;
}

static const char *encrypt_secret(redwax_tool_t *r, const char *what, const char *file,
        const char *secret, int min, int max, apr_pool_t *pool)
{


    /*
     * Obtain a secret to encrypt a key.
     *
     * Secret file specified and secret file exists, use that secret.
     *
     * Secret file specified and secret file does not exist, generate
     * a random secret and write the secret to the file.
     *
     * No secret file specified, tell openssl to ask for the secret from
     * the ui twice.
     */

    if (secret) {

        apr_file_t *sfile;
        apr_status_t status;

        status = apr_file_open(&sfile, secret, APR_FOPEN_READ, APR_FPROT_OS_DEFAULT, pool);

        if (APR_SUCCESS == status) {

            char *buf = apr_palloc(pool, max + 2);
            char *lf;

#if HAVE_APR_CRYPTO_CLEAR
            apr_crypto_clear(pool, buf, max + 2);
#endif

            status = apr_file_gets(buf, max + 2, sfile);

            if (APR_SUCCESS != status) {
                redwax_print_error(r,
                        "Could not read '%s': %pm\n", secret, &status);
                return NULL;
            }

            lf = strrchr(buf, '\n');
            if (lf) {
                *lf = 0;
            }

            return buf;

        }
        else if (APR_ENOENT == status) {

        }
        else {
            redwax_print_error(r,
                    "Could not open '%s': %pm\n", secret, &status);
            return NULL;
        }

    }

    else {

        UI *ui;
        char *prompt = NULL;

        char *buf = apr_palloc(pool, max + 2);
        char *buff = apr_palloc(pool, max + 2);

#if HAVE_APR_CRYPTO_CLEAR
        apr_crypto_clear(pool, buf, max + 2);
        apr_crypto_clear(pool, buff, max + 2);
#endif

        ui = UI_new();
        if (ui == NULL) {
            return NULL;
        }

        apr_pool_cleanup_register(pool, ui, cleanup_ui,
                apr_pool_cleanup_null);

        if ((prompt = UI_construct_prompt(ui, what,
                                          file)) == NULL) {
            return NULL;
        }

        apr_pool_cleanup_register(pool, prompt, cleanup_alloc,
                apr_pool_cleanup_null);

        if (UI_add_input_string(ui, prompt, 0, buf, min, max) < 0
            || (UI_add_verify_string(ui, prompt, 0, buff, min,
                                        max, buf) < 0)) {

            return NULL;
        }

        if (!UI_process(ui)) {
            return buf;
        }
        else {
            redwax_print_error(r,
                    "Could not read %s. Must be between %d and %d characters.\n",
                    what, min, max);
        }
    }

    return NULL;
}

apr_status_t decrypt_secret(redwax_tool_t *r, const char *file,
        const char *secret)
{

    /*
     * Obtain a secret to decrypt a key.
     *
     * Secret file specified and secret file exists, use that secret.
     *
     * Secret file specified and secret file does not exist, bail out
     * with an error.
     *
     * No secret file specified, and we are on the console, ask for
     * the secret from the console once.
     *
     * No secret file specified, and we are not on the console. Bail
     * out with an error.
     */

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_process_pkcs12_out(redwax_tool_t *r,
        const char *file, const char *secret)
{
    PKCS12 *p12 = NULL;
    X509 *x = NULL;
    STACK_OF(X509) *xis = NULL;

    const char *pass;
    const char *name = NULL;
    EVP_PKEY *pkey = NULL;
    int iter = PKCS12_DEFAULT_ITER, mac_iter = PKCS12_DEFAULT_ITER;
    int key_pbe = NID_pbe_WithSHA1And3_Key_TripleDES_CBC;
    int cert_pbe = NID_pbe_WithSHA1And3_Key_TripleDES_CBC;
    int keytype = 0;

    BIO *bio;
    int i;

    if (!strcmp(file, "-")) {
        if ((bio = BIO_new_fp(stdout, BIO_NOCLOSE)) == NULL) {
            redwax_openssl_print_errors(r);
            return APR_ENOMEM;
        }
    }
    else if ((bio = BIO_new(BIO_s_file())) == NULL) {
        redwax_openssl_print_errors(r);
        return APR_ENOMEM;
    }
    else if (BIO_write_filename(bio, (char *)file) <= 0) {
        redwax_openssl_print_errors(r);
        BIO_free(bio);
        return APR_ENOENT;
    }

    apr_pool_cleanup_register(r->pool, bio, cleanup_bio,
            apr_pool_cleanup_null);

    xis = sk_X509_new_null();

    apr_pool_cleanup_register(r->pool, xis, cleanup_sk_x509,
            apr_pool_cleanup_null);

    if (r->cert_out) {
        if (!r->certs_out->nelts) {
            redwax_print_error(r, "Warning: no certificate present, skipping pkcs12.\n");
            return APR_ENOENT;
        }
        else if (r->certs_out->nelts == 1) {

            const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->certs_out,
                    0, const redwax_certificate_t);

            const unsigned char *der = cert->der;

            x = d2i_X509(NULL, &der, cert->len);

            if (!x) {
                redwax_openssl_print_errors(r);
                return APR_ENOENT;
            }

            redwax_print_error(r, "pkcs12-out: certificate: %s\n",
                    redwax_openssl_name(r->pool, X509_get_subject_name(x)));

            apr_pool_cleanup_register(r->pool, x, cleanup_x509,
                    apr_pool_cleanup_null);

        }
        else {
            redwax_print_error(r, "Warning: more than one certificate present (%d), skipping pkcs12.\n",
                    r->certs_out->nelts);
            return APR_ENOENT;
        }
    }

    if (r->chain_out) {
        for (i = 0; i < r->intermediates_out->nelts; i++)
        {
            const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->intermediates_out,
                    i, const redwax_certificate_t);

            const unsigned char *der = cert->der;

            X509 *xi = d2i_X509(NULL, &der, cert->len);

            if (!xi) {
                redwax_openssl_print_errors(r);
                X509_free(x);
                return APR_ENOENT;
            }

            redwax_print_error(r, "pkcs12-out: intermediate: %s\n",
                    redwax_openssl_name(r->pool, X509_get_subject_name(xi)));

            sk_X509_push(xis, xi);

        }
    }

    if (r->root_out || r->trust_out) {
        for (i = 0; i < r->trusted_out->nelts; i++)
        {
            const redwax_certificate_t *cert = &APR_ARRAY_IDX(r->trusted_out, i, const redwax_certificate_t);

            const unsigned char *der = cert->der;

            X509 *xi = d2i_X509_AUX(NULL, &der, cert->len);

            if (!xi) {
                redwax_openssl_print_errors(r);
                X509_free(x);
                return APR_ENOENT;
            }

            redwax_print_error(r, "pkcs12-out: trusted: %s\n",
                    redwax_openssl_name(r->pool, X509_get_subject_name(xi)));

            sk_X509_push(xis, xi);

        }
    }

    if (r->key_out) {
        if (!r->keys_out->nelts) {
            redwax_print_error(r, "Warning: no key present, skipping pkcs12.\n");
            return APR_ENOENT;
        }
        else if (r->keys_out->nelts == 1) {

            const redwax_key_t *key = &APR_ARRAY_IDX(r->keys_out,
                    0, const redwax_key_t);

            const unsigned char *der = key->der;

            pkey = d2i_AutoPrivateKey(NULL, &der, key->len);

            if (!pkey) {
                redwax_openssl_print_errors(r);
                return APR_ENOENT;
            }

#if HAVE_EVP_PKEY_GET0_DESCRIPTION
            redwax_print_error(r, "pkcs12-out: private key: %s\n",
                    EVP_PKEY_get0_description(pkey));
#else
            redwax_print_error(r, "pkcs12-out: private key\n");
#endif

            apr_pool_cleanup_register(r->pool, pkey, cleanup_evp_pkey,
                    apr_pool_cleanup_null);

        }
        else {
            redwax_print_error(r, "Warning: more than one key present (%d), skipping.\n",
                    r->keys_out->nelts);
            return APR_ENOENT;
        }
    }

    pass = encrypt_secret(r, "PKCS12 export passphrase", file, secret,
            REDWAX_PKCS12_MIN, REDWAX_PKCS12_MAX, r->pool);
    if (!pass) {
        return APR_ENOENT;
    }

    if (r->label_out) {

        name = r->label_out;
    }
    else if (x) {

        X509_NAME *subject = X509_get_subject_name(x);
        int lastpos = -1;

        if (subject) {
            lastpos = X509_NAME_get_index_by_NID(subject, NID_commonName, -1);
        }

        if (lastpos > -1) {

            X509_NAME_ENTRY *e;
            ASN1_STRING *cn;
            BIO *lbio;
            char *buf = NULL;
            int len = 0;

            e = X509_NAME_get_entry(subject, lastpos);
            cn = X509_NAME_ENTRY_get_data(e);

            if ((lbio = BIO_new(BIO_s_mem())) == NULL) {
                return 0;
            }

            ASN1_STRING_print_ex(lbio, cn, ASN1_STRFLGS_ESC_2253 |
                    ASN1_STRFLGS_ESC_CTRL | ASN1_STRFLGS_UTF8_CONVERT);

            len = BIO_get_mem_data(lbio, &buf);

            name = apr_psprintf(r->pool, "%.*s", len, buf);

            BIO_free(lbio);

        }

    }

    p12 = PKCS12_create((char *)pass, (char *)name, pkey, x, xis,
            key_pbe, cert_pbe, iter, mac_iter, keytype);

    if (!p12) {
        redwax_openssl_print_errors(r);
        return APR_ENOENT;
    }

    else {

        apr_pool_cleanup_register(r->pool, p12, cleanup_pkcs12,
                apr_pool_cleanup_null);

        i2d_PKCS12_bio(bio, p12);
    }

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_search_chain(redwax_tool_t *r,
        const redwax_certificate_t *cert)
{
    redwax_certificate_t *ncert;
    int j;

    const unsigned char *der = cert->der;

    X509 *x = d2i_X509(NULL, &der, cert->len);
    if (x) {

        BIO *bio_err;
        bio_err = BIO_new_fp(stderr, BIO_NOCLOSE);

        BIO_printf(bio_err, "search-filter: ");
        X509_NAME_print_ex(bio_err,
                        X509_get_subject_name(x),
                        0, XN_FLAG_RFC2253);
        BIO_printf(bio_err, "\n");
        BIO_free(bio_err);

        if (!X509_NAME_cmp(X509_get_issuer_name(x),
                X509_get_subject_name(x))) {

            X509_free(x);

            return APR_ENOENT;
        }

        for (j = 0; j < sk_X509_num(trusted_index); j++) {
             X509 *xi = sk_X509_value(trusted_index, j);

             if (X509_check_issued(xi, x) == X509_V_OK) {

                  redwax_certificate_t *chain =
                          X509_get_ex_data(xi, redwax_get_x509_index());

                  /* detect loops */

                  ncert = apr_array_push(r->trusted_out);

                  ncert->header = chain->header;
                  ncert->der = chain->der;
                  ncert->len = chain->len;

                  if (X509_NAME_cmp(X509_get_issuer_name(x),
                          X509_get_subject_name(x))) {

                      rt_run_search_chain(r, chain);
                  }
             }

         }

        for (j = 0; j < sk_X509_num(chain_index); j++) {
             X509 *xi = sk_X509_value(chain_index, j);

             if (X509_check_issued(xi, x) == X509_V_OK) {

                  redwax_certificate_t *chain =
                          X509_get_ex_data(xi, redwax_get_x509_index());

                  /* detect loops / depth */

                  ncert = apr_array_push(r->intermediates_out);

                  ncert->header = chain->header;
                  ncert->der = chain->der;
                  ncert->len = chain->len;

                  if (X509_NAME_cmp(X509_get_issuer_name(x),
                          X509_get_subject_name(x))) {

                      rt_run_search_chain(r, chain);
                  }
             }

         }

        X509_free(x);
    }
    else {
        return APR_EINVAL;
    }

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_search_key(redwax_tool_t *r,
        const redwax_certificate_t *cert)
{

    redwax_key_t *key = apr_hash_get(r->keys_index,
            cert->common.subjectpublickeyinfo_der,
            cert->common.subjectpublickeyinfo_len);

    if (key) {

        redwax_key_t *nkey = apr_array_push(r->keys_out);

        nkey->header = key->header;
        nkey->der = key->der;
        nkey->len = key->len;

    }
    else {
        return APR_ENOENT;
    }

    return APR_SUCCESS;
}

static int redwax_openssl_compare_certificate(redwax_tool_t *r,
        const redwax_certificate_t *c1, const redwax_certificate_t *c2)
{
    X509 *x1, *x2;
    const ASN1_TIME *a1, *a2, *b1, *b2;
    const unsigned char *der;
    int bc1, bc2, ac1, ac2;
    int pday, psec, diff;

    /* a present certificate always beats an absent certificate */

    if ((!c1 || !c1->len) && (!c2 || !c2->len)) {
        return SAME;
    }
    if ((!c1 || !c1->len)) {
        return RIGHT;
    }
    if ((!c2 || !c2->len)) {
        return LEFT;
    }

    if (c1->len == c2->len && !memcmp(c1->der, c2->der, c1->len)) {
        return SAME;
    }

    /* a parseable certificate always beats an unreadable certificate */

    der = c1->der;
    x1 = d2i_X509(NULL, &der, c1->len);

    der = c2->der;
    x2 = d2i_X509(NULL, &der, c2->len);

    if (!x1 && !x2) {
        return SAME;
    }
    if (!x1) {
        X509_free(x2);
        return RIGHT;
    }
    if (!x2) {
        X509_free(x1);
        return LEFT;
    }

    /* a valid certificate always beats a not-yet-valid certificate */

    b1 = X509_get0_notBefore(x1);
    b2 = X509_get0_notBefore(x2);

    bc1 = b1 ? X509_cmp_time(b1, r->now) : -1;
    bc2 = b2 ? X509_cmp_time(b2, r->now) : -1;

    if (bc1 == bc2) {
        /* we have a tie */
    }
    else if (bc1 > -1) {
        X509_free(x1);
        X509_free(x2);
        return RIGHT;
    }
    else if (bc2 > -1) {
        X509_free(x1);
        X509_free(x2);
        return LEFT;
    }

    /* a valid certificate always beats a expired certificate */

    a1 = X509_get0_notAfter(x1);
    a2 = X509_get0_notAfter(x2);

    ac1 = a1 ? X509_cmp_time(a1, r->now) : 1;
    ac2 = a2 ? X509_cmp_time(a2, r->now) : 1;

    if (ac1 == ac2) {
        /* we have a tie */
    }
    else if (ac1 < 1) {
        X509_free(x1);
        X509_free(x2);
        return RIGHT;
    }
    else if (ac2 < 1) {
        X509_free(x1);
        X509_free(x2);
        return LEFT;
    }

    /* the longest validity beats the shortest validity */

    if (ASN1_TIME_diff(&pday, &psec, a1, a2)) {
        if (!pday && !psec) {
            /* still a tie */
        }
        else if (pday < 0 || psec < 0) {
            X509_free(x1);
            X509_free(x2);
            return LEFT;
        }
        else if (pday > 0 || psec > 0) {
            X509_free(x1);
            X509_free(x2);
            return RIGHT;
        }
    }

    /* we still have a tie, choose one cert, but deterministically */
    diff = X509_cmp(x1, x2);

    if (diff < 0) {
        X509_free(x1);
        X509_free(x2);
        return LEFT;
    }
    else if (diff > 0) {
        X509_free(x1);
        X509_free(x2);
        return RIGHT;
    }

    /* after all this, they're the same */

    X509_free(x1);
    X509_free(x2);

    return SAME;
}

static apr_status_t redwax_openssl_normalise_key(redwax_tool_t *r,
        redwax_key_t *key, int index)
{
    /*
     * DER encoded key is present, but the components are not.
     */
    if (key->der && !key->rsa) {

        PKCS8_PRIV_KEY_INFO *p8inf = NULL;
        EVP_PKEY *pkey = NULL;
        X509_PUBKEY *pub = NULL;

        const unsigned char *der;

        der = key->der;

        p8inf = d2i_PKCS8_PRIV_KEY_INFO(NULL, &der, key->len);
        if (!p8inf) {
            /* could not unpack, let someone else try */
            return DECLINED;
        }

        apr_pool_cleanup_register(key->pool, p8inf, cleanup_p8inf,
                apr_pool_cleanup_null);

        pkey = EVP_PKCS82PKEY(p8inf);
        if (!pkey) {
            /* could not convert, let someone else try */
            return DECLINED;
        }

        apr_pool_cleanup_register(r->pool, pkey, cleanup_evp_pkey,
                apr_pool_cleanup_null);

        if (X509_PUBKEY_set(&pub, pkey)) {

            unsigned char *der;

            key->common.subjectpublickeyinfo_len = i2d_X509_PUBKEY(pub, NULL);
            key->common.subjectpublickeyinfo_der = der = apr_palloc(r->pool,
                    key->common.subjectpublickeyinfo_len);
            i2d_X509_PUBKEY(pub, &der);

            /* index the key */
            if (index) {

                if (!apr_hash_get(r->keys_index,
                        key->common.subjectpublickeyinfo_der,
                        key->common.subjectpublickeyinfo_len)) {
                    apr_hash_set(r->keys_index,
                            key->common.subjectpublickeyinfo_der,
                            key->common.subjectpublickeyinfo_len, key);
                }

            }

        }

        switch(EVP_PKEY_base_id(pkey)) {
        case EVP_PKEY_RSA: {
            /* id is the sha1 hash of the modulus */
            unsigned char digest[EVP_MAX_MD_SIZE];

            unsigned int len;

#if HAVE_EVP_PKEY_GET_BN_PARAM
            BIGNUM *n = NULL;
            BIGNUM *e = NULL;
            BIGNUM *d = NULL;
            BIGNUM *p = NULL;
            BIGNUM *q = NULL;
            BIGNUM *dmp1 = NULL;
            BIGNUM *dmq1 = NULL;
            BIGNUM *iqmp = NULL;

            EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_N, &n);
            EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_E, &e);
            EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_D, &d);
            EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_FACTOR1, &p);
            EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_FACTOR2, &q);
            EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_EXPONENT1, &dmp1);
            EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_EXPONENT2, &dmq1);
            EVP_PKEY_get_bn_param(pkey, OSSL_PKEY_PARAM_RSA_COEFFICIENT, &iqmp);
#else
            RSA *rsa = EVP_PKEY_get1_RSA(pkey);

            const BIGNUM *n = RSA_get0_n(rsa);
            const BIGNUM *e = RSA_get0_e(rsa);
            const BIGNUM *d = RSA_get0_d(rsa);
            const BIGNUM *p = RSA_get0_p(rsa);
            const BIGNUM *q = RSA_get0_q(rsa);
            const BIGNUM *dmp1 = RSA_get0_dmp1(rsa);
            const BIGNUM *dmq1 = RSA_get0_dmq1(rsa);
            const BIGNUM *iqmp = RSA_get0_iqmp(rsa);
#endif

            key->rsa = apr_pcalloc(key->pool, sizeof(redwax_key_rsa_t));

            /* public */
            key->rsa->modulus_len = BN_num_bytes(n);
            key->rsa->modulus = apr_palloc(key->pool,
                    key->rsa->modulus_len);
            BN_bn2bin(n, key->rsa->modulus);

            key->rsa->public_exponent_len = BN_num_bytes(e);
            key->rsa->public_exponent = apr_palloc(key->pool,
                    key->rsa->public_exponent_len);
            BN_bn2bin(e, key->rsa->public_exponent);

            /* private */
            key->rsa->private_exponent_len = BN_num_bytes(d);
            key->rsa->private_exponent = apr_palloc(key->pool,
                    key->rsa->private_exponent_len);
            BN_bn2bin(d, key->rsa->private_exponent);

            key->rsa->prime_1_len = BN_num_bytes(p);
            key->rsa->prime_1 = apr_palloc(key->pool,
                    key->rsa->prime_1_len);
            BN_bn2bin(p, key->rsa->prime_1);

            key->rsa->prime_2_len = BN_num_bytes(q);
            key->rsa->prime_2 = apr_palloc(key->pool,
                    key->rsa->prime_2_len);
            BN_bn2bin(q, key->rsa->prime_2);

            key->rsa->exponent_1_len = BN_num_bytes(dmp1);
            key->rsa->exponent_1 = apr_palloc(key->pool,
                    key->rsa->exponent_1_len);
            BN_bn2bin(dmp1, key->rsa->exponent_1);

            key->rsa->exponent_2_len = BN_num_bytes(dmq1);
            key->rsa->exponent_2 = apr_palloc(key->pool,
                    key->rsa->exponent_2_len);
            BN_bn2bin(dmq1, key->rsa->exponent_2);

            key->rsa->coefficient_len = BN_num_bytes(iqmp);
            key->rsa->coefficient = apr_palloc(key->pool,
                    key->rsa->coefficient_len);
            BN_bn2bin(iqmp, key->rsa->coefficient);



            /* ID is a SHA1 hash of the modulus */
            EVP_Digest(key->rsa->modulus, key->rsa->modulus_len, digest, &len,
                    EVP_sha1(), NULL);

            key->common.id_der = apr_pmemdup(key->pool, digest, len);
            key->common.id_len = len;

#if HAVE_EVP_PKEY_GET_BN_PARAM
            BN_clear_free(n);
            BN_clear_free(e);
            BN_clear_free(d);
            BN_clear_free(p);
            BN_clear_free(q);
            BN_clear_free(dmp1);
            BN_clear_free(dmq1);
            BN_clear_free(iqmp);
#else
            RSA_free(rsa);
#endif

            break;
        }
#if 0
        case EVP_PKEY_DSA:
            break;
        case EVP_PKEY_DH:
            break;
        case EVP_PKEY_EC:
            break;
#endif
        }

    }

    /*
     * RSA components are present, but the DER encoded key is not.
     */
    else if (key->rsa && !key->der) {

    }

    return APR_SUCCESS;
}

static apr_status_t redwax_openssl_normalise_certificate(redwax_tool_t *r,
        redwax_certificate_t *cert, int index)
{

    /*
     * DER encoded certificate is present, but the components are not.
     */
    if (cert->der && !cert->x509) {

        X509 *x;
        X509_NAME *name;
        X509_PUBKEY *pub;
        ASN1_INTEGER *integer;
        ASN1_OCTET_STRING *skid;

        const unsigned char *der;

        redwax_certificate_x509_t *x509;

        der = cert->der;

        x = d2i_X509(NULL, &der, cert->len);
        if (!x) {
            /* could not unpack, let someone else try */
            return DECLINED;
        }

        if (REDWAX_CERTIFICATE_UNSPECIFIED == cert->common.category) {

            if (X509_check_ca(x)) {

                if (X509_get_extension_flags(x) & EXFLAG_SI) {
                    cert->common.category = REDWAX_CERTIFICATE_ROOT;
                }
                else {
                    cert->common.category = REDWAX_CERTIFICATE_INTERMEDIATE;
                }
            }
            else {
                cert->common.category = REDWAX_CERTIFICATE_END_ENTITY;
            }
        }

        /* handle indexing */
        if (index) {

            switch (cert->common.category) {
            case REDWAX_CERTIFICATE_END_ENTITY: {

                apr_hash_t *index;
                GENERAL_NAMES *gens = NULL;
                int i;

                /* index */
                X509_set_ex_data(x, redwax_get_x509_index(), (void *)cert);
                sk_X509_push(cert_index, x);
                X509_up_ref(x);

                gens = X509_get_ext_d2i(x, NID_subject_alt_name, NULL, NULL);
                if (gens) {
                    for (i = 0; i < sk_GENERAL_NAME_num(gens); i++) {
                        GENERAL_NAME *gen;
                        ASN1_STRING *cstr;
                        int astrlen;
                        unsigned char *astr;

                        gen = sk_GENERAL_NAME_value(gens, i);
                        switch (gen->type) {
                        case GEN_DNS: {
                            cstr = gen->d.dNSName;
                            index = r->hostnames_index;
                            break;
                        }
                        case GEN_EMAIL: {
                            cstr = gen->d.rfc822Name;
                            index = r->emails_index;
                            break;
                        }
                        case GEN_IPADD: {
                            cstr = gen->d.iPAddress;
                            index = r->ips_index;
                            break;
                        }
                        default:
                            continue;
                        }

                        astrlen = ASN1_STRING_to_UTF8(&astr, cstr);
                        if (astr) {

                            apr_pool_cleanup_register(r->pool, astr, cleanup_alloc,
                                    apr_pool_cleanup_null);

                            apr_hash_set(index, astr, astrlen,
                                    apr_pmemdup(r->pool, astr, astrlen));

                        }
                    }
                    GENERAL_NAMES_free(gens);
                }

                break;
            }
            case REDWAX_CERTIFICATE_INTERMEDIATE:

                X509_set_ex_data(x, redwax_get_x509_index(), (void *)cert);
                sk_X509_push(chain_index, x);
                X509_up_ref(x);

                break;
            case REDWAX_CERTIFICATE_ROOT:
            case REDWAX_CERTIFICATE_TRUSTED:

                X509_set_ex_data(x, redwax_get_x509_index(), (void *)cert);
                sk_X509_push(trusted_index, x);
                X509_up_ref(x);

                break;
            default:
                break;
            }

        }

        pub = X509_get_X509_PUBKEY(x);
        if (pub) {

            unsigned char *der;

            cert->common.subjectpublickeyinfo_len = i2d_X509_PUBKEY(pub, NULL);
            cert->common.subjectpublickeyinfo_der = der = apr_palloc(r->pool,
                    cert->common.subjectpublickeyinfo_len);
            i2d_X509_PUBKEY(pub, &der);
        }

        cert->common.subject = redwax_openssl_name(cert->pool,
                X509_get_subject_name(x));

        x509 = cert->x509 = apr_pcalloc(cert->pool,
                sizeof(redwax_certificate_x509_t));

        name = X509_get_subject_name(x);
        if (name) {

            unsigned char *der;

            x509->subject_len = i2d_X509_NAME(name, NULL);
            x509->subject_der = der = apr_palloc(r->pool, x509->subject_len);
            i2d_X509_NAME(name, &der);
        }

        skid = X509_get_ext_d2i(x, NID_subject_key_identifier, NULL, NULL);
        if (skid) {
            x509->id_len = skid->length;
            x509->id_der = apr_pmemdup(cert->pool, skid->data, skid->length);
        }
        else if (pub) {
            EVP_PKEY *pkey = X509_PUBKEY_get(pub);

            switch(EVP_PKEY_base_id(pkey)) {
            case EVP_PKEY_RSA: {
                /* id is the sha1 hash of the modulus */
                unsigned char digest[EVP_MAX_MD_SIZE];
                RSA *rsa = EVP_PKEY_get1_RSA(pkey);
                const BIGNUM *n = RSA_get0_n(rsa);
                unsigned char *buf = apr_palloc(cert->pool, BN_num_bytes(n));
                unsigned int len;
                BN_bn2bin(n, buf);
                EVP_Digest(buf, BN_num_bytes(n), digest, &len, EVP_sha1(), NULL);
                x509->id_der = apr_pmemdup(cert->pool, digest, len);
                x509->id_len = len;
                RSA_free(rsa);
                break;
            }
#if 0
            case EVP_PKEY_DSA:
                break;
            case EVP_PKEY_DH:
                break;
            case EVP_PKEY_EC:
                break;
#endif
            }

            EVP_PKEY_free(pkey);
        }

        name = X509_get_issuer_name(x);
        if (name) {

            unsigned char *der;

            x509->issuer_len = i2d_X509_NAME(name, NULL);
            x509->issuer_der = der = apr_palloc(r->pool, x509->issuer_len);
            i2d_X509_NAME(name, &der);
        }

        integer = X509_get_serialNumber(x);
        if (integer) {

            unsigned char *der;

            x509->serial_len = i2d_ASN1_INTEGER(integer, NULL);
            x509->serial_der = der = apr_palloc(r->pool, x509->serial_len);
            i2d_ASN1_INTEGER(integer, &der);
        }

        X509_free(x);

    }

    return APR_SUCCESS;
}

void redwax_add_default_openssl_hooks()
{
    rt_hook_initialise(redwax_openssl_initialise, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_complete_verify_param(redwax_openssl_complete_verify_param, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_set_verify_param(redwax_openssl_set_verify_param, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_process_pem_in(redwax_openssl_process_pem_in, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_complete_filter(redwax_openssl_complete_filter_verify, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_process_filter(redwax_openssl_process_filter_verify, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_complete_filter(redwax_openssl_complete_filter_search, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_process_filter(redwax_openssl_process_filter_search, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_process_pem_out(redwax_openssl_process_pem_out, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_process_pkcs12_out(redwax_openssl_process_pkcs12_out, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_search_chain(redwax_openssl_search_chain, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_search_key(redwax_openssl_search_key, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_compare_certificate(redwax_openssl_compare_certificate, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_normalise_key(redwax_openssl_normalise_key, NULL, NULL, APR_HOOK_MIDDLE);
    rt_hook_normalise_certificate(redwax_openssl_normalise_certificate, NULL, NULL, APR_HOOK_MIDDLE);
}

#endif
